#include "scp_aml_i2c.h"
#include "task_apis.h"
#include "secure_apb.h"

#define MIN(x,y)            				(((x) < (y)) ? (x) : (y))

#define AML_I2C_SLAVE_ADDR_MASK        		0xFF
#define AML_I2C_SLAVE_ADDR_MASK_7BIT   		0x7F

#define I2C_M_TEN                        	0x0010    /* this is a ten bit chip address */
#define I2C_M_RD                        	0x0001    /* read data, from slave to master */
#define I2C_M_NOSTART                    	0x4000    /* if I2C_FUNC_PROTOCOL_MANGLING */
#define I2C_M_REV_DIR_ADDR                	0x2000    /* if I2C_FUNC_PROTOCOL_MANGLING */
#define I2C_M_IGNORE_NAK                	0x1000    /* if I2C_FUNC_PROTOCOL_MANGLING */
#define I2C_M_NO_RD_ACK                    	0x0800    /* if I2C_FUNC_PROTOCOL_MANGLING */
#define I2C_M_RECV_LEN                    	0x0400    /* length will be first received byte */

void scp_aml_i2c_init(struct scp_aml_i2c *i2c, unsigned int reg_ctrl_address)
{
    i2c->master_i2c_speed = 100000;
    i2c->wait_count = 1000000;
    i2c->wait_ack_interval = 5;
    i2c->wait_read_interval = 5;
    i2c->wait_xfer_interval = 5;

    i2c->master_regs = (struct scp_aml_i2c_reg_master *)reg_ctrl_address;
}

static void scp_aml_i2c_clear_token_list(struct scp_aml_i2c *i2c)
{
    int i;
    i2c->master_regs->i2c_token_list_0 = 0;
    i2c->master_regs->i2c_token_list_1 = 0;
    for (i = 0; i < AML_I2C_MAX_TOKENS; i++) {
        i2c->token_tag[i] = TOKEN_END;
    }
}

static void scp_aml_i2c_set_token_list(struct scp_aml_i2c *i2c)
{
    int i;
    unsigned int token_reg = 0;

    for (i = 0; i < AML_I2C_MAX_TOKENS; i++)
        token_reg |= i2c->token_tag[i] << (i * 4);

    i2c->master_regs->i2c_token_list_0 = token_reg;
}

static int scp_aml_i2c_check_error(struct scp_aml_i2c *i2c)
{
    struct scp_aml_i2c_reg_ctrl* ctrl;
    ctrl = (struct scp_aml_i2c_reg_ctrl *)&(i2c->master_regs->i2c_ctrl);

    if (ctrl->error) {
        return -1;
    }
    return 0;
}

static int scp_aml_i2c_wait_ack(struct scp_aml_i2c *i2c)
{
    int i;
    struct scp_aml_i2c_reg_ctrl *ctrl;

    for (i = 0; i < i2c->wait_count; i++) {
        _udelay(i2c->wait_ack_interval);
        ctrl = (struct scp_aml_i2c_reg_ctrl*)&(i2c->master_regs->i2c_ctrl);
        if (I2C_IDLE == ctrl->status) {
            return scp_aml_i2c_check_error(i2c);
        }
    }

    return -1;
}

static void scp_aml_i2c_fill_data(struct scp_aml_i2c *i2c, unsigned char *buf, unsigned len)
{
    int i;
    unsigned int wdata0 = 0;
    unsigned int wdata1 = 0;

    for (i = 0; i < MIN(len, AML_I2C_MAX_TOKENS >> 1); i++) {
        wdata0 |= (*buf++) << (i * 8);
    }

    for (; i < MIN(len, AML_I2C_MAX_TOKENS); i++) {
        wdata1 |= (*buf++) << ((i - (AML_I2C_MAX_TOKENS >> 1)) * 8);
    }

    i2c->master_regs->i2c_token_wdata_0 = wdata0;
    i2c->master_regs->i2c_token_wdata_1 = wdata1;
}

static void scp_aml_i2c_start_token_xfer(struct scp_aml_i2c *i2c)
{
    /* re-start the list processor with a new list
     * after a previous list has been exhausted
     * */
    i2c->master_regs->i2c_ctrl &= ~1;  /*clear*/
    i2c->master_regs->i2c_ctrl |= 1;   /*set*/

    _udelay(i2c->wait_xfer_interval);
}

static void scp_aml_i2c_set_address(struct scp_aml_i2c *i2c, unsigned int addr)
{
    i2c->cur_slave_addr = addr & AML_I2C_SLAVE_ADDR_MASK_7BIT;
    i2c->master_regs->i2c_slave_addr &= (~AML_I2C_SLAVE_ADDR_MASK);
    i2c->master_regs->i2c_slave_addr |= (i2c->cur_slave_addr << 1);
}

static void scp_aml_i2c_stop(struct scp_aml_i2c *i2c)
{
    struct scp_aml_i2c_reg_ctrl *ctrl;
    ctrl = (struct scp_aml_i2c_reg_ctrl *)&(i2c->master_regs->i2c_ctrl);
    if (!ctrl->error) {
        scp_aml_i2c_clear_token_list(i2c);
        i2c->token_tag[0] = TOKEN_STOP;
        scp_aml_i2c_set_token_list(i2c);
        scp_aml_i2c_start_token_xfer(i2c);
        scp_aml_i2c_wait_ack(i2c);
        i2c->master_regs->i2c_ctrl &= ~1;  /*clear*/
    }
}

static int scp_aml_i2c_write(struct scp_aml_i2c *i2c, unsigned char *buf, unsigned int len)
{
    int i;
    int ret;
    unsigned wr_len;
    int tagnum = 0;

    scp_aml_i2c_clear_token_list(i2c);

    if (!(i2c->msg_flags & I2C_M_NOSTART)) {
        i2c->token_tag[tagnum++] = TOKEN_START;
        i2c->token_tag[tagnum++] = TOKEN_SLAVE_ADDR_WRITE;
    }
    while (len) {
        wr_len = MIN(len, AML_I2C_MAX_TOKENS - tagnum);
        for (i=0; i < wr_len; i++) {
            i2c->token_tag[tagnum++] = TOKEN_DATA;
        }

        scp_aml_i2c_set_token_list(i2c);

        scp_aml_i2c_fill_data(i2c, buf, wr_len);

        scp_aml_i2c_start_token_xfer(i2c);

        len -= wr_len;
        buf += wr_len;
        tagnum = 0;

        ret = scp_aml_i2c_wait_ack(i2c);
        if (ret < 0) {
            i2c->master_regs->i2c_ctrl &= ~1;  /*clear*/
            return ret;
        }

        scp_aml_i2c_clear_token_list(i2c);
    }
    i2c->master_regs->i2c_ctrl &= ~1;  /*clear*/
    return 0;
}

int scp_aml_i2c_send_buf(struct scp_aml_i2c *i2c, unsigned int addr, unsigned char *buf, unsigned int len)
{
    int ret = 0;

    scp_aml_i2c_set_address(i2c, addr);
    ret = scp_aml_i2c_write(i2c, buf, len);
    scp_aml_i2c_stop(i2c);

    return ret;
}
