#include <common.h>
#include <i2c.h>
#include <aml_i2c.h>
#include <power/power_delivery.h>
#include <power/husb238.h>

#define HUSB238_REG__PD_STATUS0		0x00

#define PD_SRC_VOLTAGE__MSK		0xF0
#define PD_SRC_VOLTAGE__OFFSET		4
#define PD_SRC_VOLTAGE__UNATTACHED	0x0
#define PD_SRC_VOLTAGE__MAX		0x6

#define PD_SRC_CURRENT__MSK		0x0F
#define PD_SRC_CURRENT__OFFSET		0

#define HUSB238_REG__SRC_PDO		0x08
#define SRC_PDO__MSK			0xF0
#define SRC_PDO__OFFSET			4

#define HUSB238_REG__GO_COMMAND		0x09
#define GO_COMMAND__SELECT_PDO		0x01

#define SRC_PDO__DETECT_MSK		0x80

#define husb238_msg(fmt, args...)	printf("[husb238]: "fmt, ##args)
#define husb238_err(fmt, args...)	printf("[husb238]: "fmt, ##args)


static struct husb238 husb238;

static const unsigned int pd_voltage[] = {
    [0x1] = 5000,
    [0x2] = 9000,
    [0x3] = 12000,
    [0x4] = 15000,
    [0x5] = 18000,
    [0x6] = 20000,
};

static const unsigned int pd_current[] = {
	500,
	700,
	1000,
	1250,
	1500,
	1750,
	2000,
	2250,
	2500,
	2750,
	3000,
	3250,
	3500,
	4000,
	4500,
	5000,
};

static int husb238_i2c_read_byte(u16 addr, u8 reg, u8 *data)
{
	int ret;

	struct i2c_msg msg[2] = {
		{
			.addr = addr,
			.flags = 0,
			.len = 1,
			.buf = &reg,
		},
		{
			.addr = addr,
			.flags = I2C_M_RD,
			.len = 1,
			.buf = data,
		},
	};
	ret = aml_i2c_xfer(msg, 2);
	if (ret != 2) {
		husb238_err("husb238: i2c read failed: %d\n", ret);
		return ret;
	}

	return 0;
}

int husb238_get_status(struct husb238 *dev, struct pd_status *status)
{
	u8 pd_src_status;
	int ret;
	unsigned pd_src_voltage, pd_src_current;

	status->is_pd = 0;
	status->voltage = 0;
	status->curr = 0;

	ret = i2c_set_bus_num(dev->i2c_busnum);
	if (ret) {
		husb238_err("set i2c bus failed: %d\n", ret);
		return -1;
	}

	ret = husb238_i2c_read_byte(dev->i2c_addr, HUSB238_REG__PD_STATUS0,
				    &pd_src_status);
	if (ret)
		return -1;

	pd_src_voltage = (pd_src_status & PD_SRC_VOLTAGE__MSK) >>
			 PD_SRC_VOLTAGE__OFFSET;
	if (pd_src_voltage > PD_SRC_VOLTAGE__MAX) {
		husb238_err("Undefined PD_SRC_VOLTAGE value: 0x%01x\n",
			    pd_src_voltage);
		return -1;
	}

	if (pd_src_voltage == PD_SRC_VOLTAGE__UNATTACHED) {
		return 0;
	}

	pd_src_current = (pd_src_status & PD_SRC_CURRENT__MSK) >>
			 PD_SRC_CURRENT__OFFSET;

        status->is_pd = 1;
        status->voltage = pd_voltage[pd_src_voltage];
	status->curr = pd_current[pd_src_current];
	return 0;
}

int do_husb238(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
	char *cmd;
	struct pd_status status;

	if (argc < 2)
		return CMD_RET_USAGE;

	cmd = argv[1];

	if (strcmp(cmd, "init") == 0) {
		if (argc < 4)
			return CMD_RET_USAGE;
		husb238.i2c_busnum = simple_strtoul(argv[2], NULL, 10);
		husb238.i2c_addr = simple_strtoul(argv[3], NULL, 16);
		husb238.inited = 1;
		husb238_msg("new busnum: %d; new addres: 0x%02x\n",
			    husb238.i2c_busnum, husb238.i2c_addr);
		return CMD_RET_SUCCESS;
	}

	if (!husb238.inited) {
		husb238_err("must call init first\n");
		return CMD_RET_FAILURE;
	}

	if (strcmp(cmd, "status") == 0) {
		ret = husb238_get_status(&husb238, &status);
		if (ret)
			return CMD_RET_FAILURE;

		husb238_msg("is_pd: %d; voltage: %d; curr: %d\n",
			    status.is_pd, status.voltage, status.curr);
		return CMD_RET_SUCCESS;
	}

	return CMD_RET_SUCCESS;
}

U_BOOT_CMD(
	husb238, CONFIG_SYS_MAXARGS, 1, do_husb238,
	"husb238",
	"init busnum address - set i2c bus number and i2c address\n"
	"husb238 status - read and print current status\n"
);
