/* SPDX-License-Identifier: GPL-2.0-only */
/*
 * Copyright (C) 2021 Yandex Inc
 *
 * Gowin led device communication protocol client api for application level.
 * Based on https://wiki.yandex-team.ru/smartdevices/station-2displayfpga
 */

#include <common.h>
#include <command.h>
#include <crc.h>
#include <malloc.h>
#include <errno.h>
#include <linux/byteorder/little_endian.h>
#include <dm.h>
#include <spi.h>
#include <gowin_led_device.h>
#include "gowin.h"

#define gowin_led_device_msg(fmt, args...) printf("[gowin_led_device msg]: "fmt, ##args)
#define gowin_led_device_err(fmt, args...) printf("[gowin_led_device err]: "fmt, ##args)

static void response2err(u8 status) {

	if (status & COMMAND_ERROR)
		gowin_led_device_err("CMD_ERROR\n");

	if (status & CRC_ERROR)
		gowin_led_device_err("CRC_ERROR\n");

	if (status & CMD_OVF_ERROR)
		gowin_led_device_err("CMD_OVF_ERROR\n");

	if (status & FRM_OVF_ERROR)
		gowin_led_device_err("FRM_OVF_ERROR\n");

	if (status & SPKT_ERROR)
		gowin_led_device_err("SPKT_ERROR\n");

	if (status & LPKT_ERROR)
		gowin_led_device_err("LPKT_ERROR\n");
}

static void pr_response(const struct gld_response *response) {
	gowin_led_device_err("led panel response:\n");
	gowin_led_device_err("fw_version = 0x%x\n", response->fw_version);
	gowin_led_device_err("frame_overflows = 0x%x\n", response->frame_overflows);
	gowin_led_device_err("crc_errors = 0x%x\n", response->crc_errors);
	gowin_led_device_err("frames_in_queue = 0x%x\n", response->frames_in_queue);
	gowin_led_device_err("last_cmd_status = 0x%x\n", response->last_cmd_status);
	gowin_led_device_err("uf_crc = 0x%x\n", response->uf_crc);
	gowin_led_device_err("crc = 0x%x\n", response->crc);
}

static void print_hex_dump(u8 *buf, int size)
{
	int idx = 0;
	int i;
	char str[512];
	int res;
	for (i = 0; i < size; i++) {
		res = snprintf(&str[idx], sizeof(str) - idx - 2, "%02X", buf[i]);
		if (res < 0)
			break;
		idx += res;
		if (idx > sizeof(str) - 2) {
			idx = sizeof(str) - 2;
			break;
		}
	}
	str[idx] = '\n';
	str[idx + 1] = '\0';
	printf(str);
}

static void parse_response(const u8 *buf, struct gld_response *response)
{
	// command
	++buf;

	response->mode = *buf++;
	response->fw_version = *buf++;

	memcpy(&response->board_rev, buf, SIZE__BOARD_REV);
	response->board_rev = le16_to_cpu(response->board_rev);
	buf += SIZE__BOARD_REV;

	memcpy(&response->uf_crc, buf, SIZE__CRC);
	response->uf_crc = le16_to_cpu(response->uf_crc);
	buf += SIZE__CRC;

	memcpy(&response->frame_overflows, buf, SIZE__FRAME_OVERFLOWS);
	response->frame_overflows = le32_to_cpu(response->frame_overflows);
	buf += SIZE__FRAME_OVERFLOWS;
	memcpy(&response->crc_errors, buf, SIZE__CRC_ERRORS);
	response->crc_errors = le32_to_cpu(response->crc_errors);
	buf += SIZE__CRC_ERRORS;

	response->frames_in_queue = *buf++;
	response->last_cmd_status = *buf++;
	// Reserved and touch fields
	buf += SIZE__RESERVED;

	memcpy(&response->crc, buf, SIZE__CRC);
	response->crc = le16_to_cpu(response->crc);
}

static int gld_get_status(struct gld_response *response, int retries,
			       struct spi_slave *slave, struct udevice *dev)
{
	u8 rx[SIZE__RESPONSE] = {0};
	int err;

	for (; retries > 0; --retries) {
		err = spi_xfer(slave, 8 * SIZE__RESPONSE, NULL, rx,
			       SPI_XFER_BEGIN | SPI_XFER_END);
		if (err) {
			gowin_led_device_err("spi transaction failed: %d (%s)\n",
				      err,  __func__);
			return err;
		}

		parse_response(rx, response);

		if(cyg_crc16(rx, SIZE__RESPONSE)) {
			gowin_led_device_err("response crc mismatch\n");
			pr_response(response);
			print_hex_dump(rx, SIZE__RESPONSE);
			mdelay(10);
			err = -EREMOTEIO;
			continue;
		}

		break;
	}

	if (response->mode != GLD_APPLICATION_MODE) {
		gowin_led_device_err(
			"application mode number mismatch: expected=0x%x, "
			"actual=0x%x\n",
			GLD_APPLICATION_MODE, response->mode);
		err = -EPROTO;
	}

	return err;
}

static int gld_write(u8 *buf, size_t size, int retries)
{
	struct gld_response response = {0};
	struct spi_slave *slave;
	struct udevice *dev;
	int err;
	u16 crc;

	crc = cyg_crc16(buf, size - SIZE__CRC);
	crc = cpu_to_be16(crc);
	memcpy(buf + (size - SIZE__CRC), &crc, SIZE__CRC);

	err = spi_get_bus_and_cs(GLD_SPI_BUS_NUMBER,
				 GLD_SPI_CS,
				 GLD_SPI_DEVICE_SPEED,
				 SPI_MODE_0, "spi_generic_drv",
				 GLD_SPI_DEVICE_NAME, &dev, &slave);
	if (err) {
		gowin_led_device_err("spi get bus failed: %d\n", err);
		return err;
	}

	slave->wordlen = 8;
	err = spi_claim_bus(slave);
	if (err) {
		gowin_led_device_err("spi claim bus failed: %d\n", err);
		goto spi_err;
	}

	for (; retries > 0; --retries) {
		err = spi_xfer(slave, 8 * size, buf, NULL,
			       SPI_XFER_BEGIN | SPI_XFER_END);
		if (err) {
			gowin_led_device_err("spi transaction failed: %d (%s)\n",
				      err,  __func__);
			return err;
		}

		err = gld_get_status(&response, retries, slave, dev);
		if (err)
			goto spi_err;

		if (response.last_cmd_status & CMD_OVF_ERROR) {
			mdelay(10);
			continue;
		}

		if (response.last_cmd_status & CRC_ERROR) {
			mdelay(10);
			continue;
		}

		break;
	}
	if (response.last_cmd_status) {
		response2err(response.last_cmd_status);
		err = -EREMOTEIO;
		goto spi_err;
	}

spi_err:
	spi_release_bus(slave);
	spi_free_slave(slave);

	return err;
}

int gld_get_board_rev(void)
{
	struct spi_slave *slave;
	struct udevice *dev;
	int err;
	struct gld_response response = {0};

	err = spi_get_bus_and_cs(GLD_SPI_BUS_NUMBER,
				 GLD_SPI_CS,
				 GLD_SPI_DEVICE_SPEED,
				 SPI_MODE_0, "spi_generic_drv",
				 GLD_SPI_DEVICE_NAME, &dev, &slave);
	if (err) {
		gowin_led_device_err("spi get bus failed: %d\n", err);
		return err;
	}

	slave->wordlen = 8;
	err = spi_claim_bus(slave);
	if (err) {
		gowin_led_device_err("spi claim bus failed: %d\n", err);
		goto spi_err;
	}

	err = gld_get_status(&response, RETRIES__GET_STATUS, slave, dev);

	spi_err:
	spi_release_bus(slave);
	spi_free_slave(slave);

	return response.board_rev;
}

int gld_set_animation_mode(u8 animation_mode)
{
	u8 cmd[SIZE__SET_ANIMATION] = {0};
	cmd[0] = CMD__SET_ANIMATION;
	cmd[1] = animation_mode;
	return gld_write(cmd, SIZE__SET_ANIMATION, RETRIES__SET_ANIMATION);
}

int gld_set_brightness(u8 brightness)
{
	u8 cmd[SIZE__SET_BRIGHTNESS];
	cmd[0] = CMD__SET_BRIGHTNESS;
	cmd[1] = brightness;
	return gld_write(cmd, SIZE__SET_BRIGHTNESS, RETRIES__SET_BRIGHTNESS);
}

int gld_show_pic(u8 *buf, size_t pic_size) {
	int err;
	u8 *cmd;

	cmd = malloc(SIZE__SHOW_PIC_WITHOUT_BUF + pic_size);
	if (!cmd) {
		gowin_led_device_err("can't alloc memory for picture\n");
		return -ENOMEM;
	}

	cmd[0] = CMD__SHOW_PIC;
	cmd[1] = 0x0; // reserved
	// not zero delay
	memset(&cmd[2], 0x1, 2);

	memcpy(&cmd[4], buf, pic_size);

	err = gld_write(cmd, SIZE__SHOW_PIC_WITHOUT_BUF + pic_size,
		        RETRIES__SHOW_PIC);
	free(cmd);

	return err;
}
