#include <common.h>
#include <asm-generic/gpio.h>

#include <gpio_long_press.h>

#define DEBOUNCE_TIMEOUT_MS (50)
#define DEBOUNCE_LOOP_COUNT (5)
#define DEBOUNCE_LOOP_ERR_LIMIT (DEBOUNCE_LOOP_COUNT / 2)

static int read_gpio(int gpio, int debounce_loop_cnt,
        int debounce_loop_err_limit, int debounce_timeout_ms)
{
    size_t err_count = 0;
    int old_value = -1;
    int delay_ms;

    delay_ms = debounce_timeout_ms / debounce_loop_cnt;

    old_value = gpio_get_value(gpio);
    if (old_value < 0) {
        printf("Error reading gpio %d\n", gpio);
        return old_value;
    }

    while (debounce_loop_cnt > 0) {
        int value = -1;

        value = gpio_get_value(gpio);
        if (value < 0) {
            printf("Error reading gpio %d\n", gpio);
            return value;
        }

        if (value != old_value) {
            err_count++;
        }

        if (err_count > debounce_loop_err_limit) {
            return -1;
        }

        old_value = value;

        mdelay(delay_ms);
        debounce_loop_cnt--;
    }

    return (err_count <= debounce_loop_err_limit) ? old_value : -1;
}

int gpio_long_press_check(int gpio, int timeout_ms,
        struct list_head *notifiers_list)
{
    struct gpio_long_press_notifier *n = NULL;
    int ret = 0;

    if (!notifiers_list || list_empty(notifiers_list)) {
        return -1;
    }

    ret = gpio_request(gpio, "gpio_long_press");
    if (ret < 0) {
        printf("Error requesting gpio %d\n", gpio);
        return ret;
    }

    ret = gpio_direction_input(gpio);
    if (ret < 0) {
        printf("Error setting input for gpio %d\n", gpio);
        goto exit_free;
    }

    while (timeout_ms > 0) {
        int timer_before = get_timer(0);
        int timer_after;

        ret = read_gpio(gpio, DEBOUNCE_LOOP_COUNT,
            DEBOUNCE_LOOP_ERR_LIMIT, DEBOUNCE_TIMEOUT_MS);

        if (ret < 0) {
            printf("read_gpio failed\n");
            goto exit_free;
        } else if (ret == 0) {
            goto exit_free;
        }

        timer_after = get_timer(0);

        timeout_ms -= (timer_after - timer_before);

        putc('.');
    }

    putc('\n');

    list_for_each_entry(n, notifiers_list, list) {
        if (n->function) {
            ret |= n->function(n->data);
        }
    }

exit_free:
    gpio_free(gpio);

    return ret;
}
