#include <common.h>
#include <pcba_id.h>
#include <crc.h>
#include <linux/ctype.h>

#define pcba_id_msg(fmt, args...)       printf("[pcba_id msg]: "fmt, ##args)
#define pcba_id_err(fmt, args...)       printf("[pcba_id err] (%s): "fmt, __func__, ##args)

#define INSIGNIFICANT_REV_PART			-1

/**
 * @brief Parse PCBA ID numeric field
 * @param buf - input string for parse
 * @param len - field size in PCBA ID string
 * @param field  - pointer to output numeric field
 * @param base - base of numeric field: 10 for decimal, 16 for hexadecimal
 * @return Size of parsed field
 */
static int pcba_id_parse_digit_field(char *buf, int len, unsigned *field, int base)
{
	char temp[128];
	memcpy(temp, buf, len);
	temp[len] = '\0';
	*field = simple_strtoul(temp, NULL, base);
	return len;
}

/**
 * @brief Parse PCBA ID string field
 * @param buf - input string for parse
 * @param len - field size in PCBA ID string
 * @param field  - output field. Field size must be (len + 1).
 *        It is necessary for storage zero-terminate symbol.
 * @return Size of parsed field
 */
static int pcba_id_parse_string_field(char *buf, int len, char *field)
{
	memcpy(field, buf, len);
	field[len] = '\0';
	return len;
}

/**
 * @brief Read buffer from EEPROM IC by i2c
 * @param i2c_address - address IC on i2c bus
 * @param buf - result buffer
 * @param len - read buffer size
 * @return CMD_RET_SUCCESS - command successfully finished, else - error
 */
static int pcba_id_read_buf_from_i2c(unsigned i2c_address, char *buf, unsigned len)
{
	int ret;
	char cmd[128];

	snprintf(cmd, sizeof(cmd), "i2c dev 0");
	ret = run_command(cmd, 0);
	if (ret != 0) {
		pcba_id_err("fail in cmd: %s; ret: %d\n", cmd, ret);
		return CMD_RET_FAILURE;
	}

	snprintf(cmd, sizeof(cmd), "i2c read %02x 00 %02x %p", i2c_address, len, buf);
	ret = run_command(cmd, 0);
	if (ret != 0) {
		pcba_id_err("fail in cmd: %s; ret: %d\n", cmd, ret);
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

/**
 * @brief Check if the string is number
 * @param s  - array of bytes
 * @return 0 - invalid string, 1 - valid string
 */
static int is_number(char *s)
{
	int i;

	for(i = 0; s[i] != '\0'; i++) {
		if(isdigit(s[i]) == 0)
			return 0;
	}
	return 1;
}

/**
 * @brief Check if char is valid
 * @param c - one char
 * @return 0 - invalid string, 1 - valid string
 */
static int is_valid_char(char c) {
	if (!((isalpha(c) && isascii(c)) || isdigit(c) || c == '.'))
		return 0;

	return 1;
}

/**
 * @brief Check if the string contains valid characters
 * @param buf - array of bytes
 * @param len - array length
 * @return 0 - invalid string, 1 - valid string
 */
static int is_valid_string(const char *buf, int len)
{
	int i;
	for(i = 0; i < len; i++) {
		if (!is_valid_char(buf[i]))
			return 0;
	}
	return 1;
}

/**
 * @brief Initialize pcba_id structure for pcba_id v1
 * @param buf - array of bytes
 * @param pcba_id - result PCBA ID struct
 * @return 0 - read bytes number
 */
static int parse_sv1(char *buf, struct pcba_id *pcba_id) {
	unsigned checksum;
	int idx = FIELD_SIZE__STRUCT_VERSION;

	idx += pcba_id_parse_string_field(&buf[idx], SV1_FIELD_SIZE__PRODUCT_OWNER,
									  pcba_id->product_owner);
	idx += pcba_id_parse_digit_field(&buf[idx], SV1_FIELD_SIZE__PRODUCT_MODEL,
									 &pcba_id->product_model, 10);
	idx += pcba_id_parse_string_field(&buf[idx], SV1_FIELD_SIZE__MANUFACTURER,
									  pcba_id->manufacturer);
	idx += pcba_id_parse_digit_field(&buf[idx], 2, &pcba_id->board_code, 10);
	idx += pcba_id_parse_digit_field(&buf[idx], 2, &pcba_id->board_revision.hi, 10);
	idx += pcba_id_parse_digit_field(&buf[idx], 2, &pcba_id->board_revision.mid, 10);
	idx += pcba_id_parse_digit_field(&buf[idx], 2, &pcba_id->board_revision.lo, 10);
	idx += pcba_id_parse_string_field(&buf[idx], FIELD_SIZE__DATE, pcba_id->date);
	idx += pcba_id_parse_digit_field(&buf[idx], 5, &pcba_id->incremental_number, 10);
	idx += pcba_id_parse_digit_field(&buf[idx], 4, &pcba_id->checksum, 16);

	checksum = cyg_crc16((unsigned char *)buf, idx - 4);
	if (checksum != pcba_id->checksum) {
		pcba_id_err("wrong checksum\n");
		pcba_id_msg("checksum calc: 0x%04X\n", checksum);
		pcba_id_msg("checksum read: 0x%04X\n", pcba_id->checksum);
		return PARSE_RES__WRONG_CRC;
	}

	return PARSE_RES__SUCCESS;
}

/**
 * @brief Initialize pcba_id structure for pcba_id v2
 * @param eeprom_buf - array of bytes
 * @param pcba_id - result PCBA ID struct
 * @return 0 - read bytes number
 */
static int parse_sv2(char *eeprom_buf, struct pcba_id *pcba_id)
{
	char *token[SV2_FIELDS_COUNT];
	char crc_buf[FIELD_SIZE__CHECKSUM + 1];
	unsigned crc;
	int tc = 0;

	/* handle checksum */
	memcpy(crc_buf, eeprom_buf + SV2_OFFSET__CHECKSUM, FIELD_SIZE__CHECKSUM);
	crc_buf[FIELD_SIZE__CHECKSUM] = '\0';
	crc = cyg_crc16((unsigned char *)eeprom_buf,
					SV2__EEPROM_SIZE - FIELD_SIZE__CHECKSUM);
	pcba_id->checksum = crc;

	if(crc != simple_strtoul(crc_buf, NULL, 16)) {
		pcba_id_err("checksum error\n");
		return PARSE_RES__WRONG_CRC;
	}

	if(!is_valid_string(eeprom_buf, SV2__EEPROM_SIZE)) {
		pcba_id_err("string contains non valid symbol\n");
		return PARSE_RES__OTHER;
	}

	/* set null terminator before CRC begin */
	eeprom_buf[SV2__EEPROM_SIZE - FIELD_SIZE__CHECKSUM] = 0;

	// struct version (2 digits)
	token[STRUCT_V] = strsep((char **)&eeprom_buf, ".");
	if (strlen(token[STRUCT_V]) != 2 || !is_number(token[STRUCT_V])) {
		pcba_id_err("struct version format error\n");
		return PARSE_RES__OTHER;
	}
	tc++;

	for(; tc < SV2_FIELDS_COUNT && token[tc - 1] != NULL; ++tc) {
		token[tc] = strsep((char **)&eeprom_buf, ".");
	}

	if (tc != SV2_FIELDS_COUNT || token[tc - 1] == NULL) {
		pcba_id_err("incorrect PCBA ID V2 format\n");
		return PARSE_RES__OTHER;
	}

	// product model (arbitrary length, digits)
	if(!is_number(token[PRODUCT_MODEL])) {
		pcba_id_err("product model format error\n");
		return PARSE_RES__OTHER;
	}

	// board code (arbitrary length, digits)
	if(!is_number(token[BOARD_CODE])) {
		pcba_id_err("board code format error\n");
		return PARSE_RES__OTHER;
	}

	// revision hi (arbitrary length, digits)
	if(!is_number(token[REVISION_HI])) {
		pcba_id_err("revision hi format error\n");
		return PARSE_RES__OTHER;
	}
	// revision mid (arbitrary length, digits)
	if(!is_number(token[REVISION_MID])) {
		pcba_id_err("revision mid format error\n");
		return PARSE_RES__OTHER;
	}
	// revision lo (arbitrary length, digits)
	if(!is_number(token[REVISION_LO])) {
		pcba_id_err("revision lo format error\n");
		return PARSE_RES__OTHER;
	}

	// date YYMMDD (6 digits)
	if (strlen(token[DATE]) != 6 || !is_number(token[DATE])) {
		pcba_id_err("date format error\n");
		return PARSE_RES__OTHER;
	}

	// serial num (6 digits)
	if (strlen(token[SERIAL]) != 6 || !is_number(token[SERIAL])) {
		pcba_id_err("serial format error\"\n  }");
		return PARSE_RES__OTHER;
	}

	strcpy(pcba_id->product_owner, token[PRODUCT_OWNER]);
	strcpy(pcba_id->manufacturer, token[MANUFACTURER]);
	strcpy(pcba_id->date, token[DATE]);
	strcpy(pcba_id->assembly, token[ASSEMBLY]);
	strcpy(pcba_id->board_revision.additional, token[REVISION_ADDITIONAL]);

	pcba_id->product_model = simple_strtoul(token[PRODUCT_MODEL], NULL, 10);
	pcba_id->struct_version = simple_strtoul(token[STRUCT_V], NULL, 10);
	pcba_id->board_code = simple_strtoul(token[BOARD_CODE], NULL, 10);
	pcba_id->incremental_number = simple_strtoul(token[SERIAL], NULL, 10);
	pcba_id->board_revision.hi = simple_strtoul(token[REVISION_HI], NULL, 10);
	pcba_id->board_revision.mid = simple_strtoul(token[REVISION_MID], NULL, 10);
	pcba_id->board_revision.lo = simple_strtoul(token[REVISION_LO], NULL, 10);
	return PARSE_RES__SUCCESS;
}

enum parse_res_t parse_eeprom(unsigned i2c_address, struct pcba_id *pcba_id)
{
	char buf[128];

	pcba_id_read_buf_from_i2c(i2c_address, buf, sizeof buf);

	pcba_id_parse_digit_field(buf, FIELD_SIZE__STRUCT_VERSION, &pcba_id->struct_version, 10);
	switch(pcba_id->struct_version) {
	case 1:
		return parse_sv1(buf, pcba_id);
	case 2:
		return parse_sv2(buf, pcba_id);
	default:
		pcba_id_err("unknown struct version\n");
		return PARSE_RES__UNKNOWN_VERSION;
	}

	return PARSE_RES__OTHER;
}
